unit MainFrm;

{
  Main form for PasCast app.
  Copyright 2015 Phil Hess.
}

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, DateUtils,
  Forms, Controls, Graphics, Dialogs, Menus, ExtCtrls, ComCtrls, StdCtrls,
  LCLProc,
  GDAL, P4G_Legend, P4G_Map, P4G_Layer, NdfdLib;

const
  ProgramName    = 'PasCast';
  ProgramVersion = '1.0.3';
  //TODO: Retrieve from .exe version resource (Windows) or Info.plist (Mac).
  // Latter requires something like PropListUtil unit
  // (see http://wiki.lazarus.freepascal.org/XDev_Toolkit).

type
  TMainForm = class(TForm)
    Legend: TP4G_Legend;
    Splitter: TSplitter;
    Map: TP4G_Map;
    ToolBar: TToolBar;
    UpdateBtn: TButton;

    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure AboutCmdClick(Sender: TObject);
    procedure HelpCmdClick(Sender: TObject);
    procedure UpdateBtnClick(Sender: TObject);
    procedure MapMouseUp(Sender: TObject; Button: TMouseButton;
                         Shift: TShiftState; X, Y: Integer);

  private
    MainMenu     : TMainMenu;
{$IFDEF DARWIN}
    AppMenu      : TMenuItem;
    AppAboutCmd  : TMenuItem;
{$ENDIF}
    HelpMenu     : TMenuItem;
    HelpHelpCmd  : TMenuItem;
{$IFNDEF DARWIN}
    HelpAboutCmd : TMenuItem;
{$ENDIF}

    Ndfd        : TNdfdLib;
    LocLayer    : TP4G_VectorLayer;
    LatLonTrans : OGRCoordinateTransformationH;

    procedure SetupMenu;
    procedure AddLayers;
    procedure SetupTransformation;
  end;

var
  MainForm: TMainForm;


implementation

{$R *.lfm}

procedure TMainForm.FormCreate(Sender: TObject);
begin
  Ndfd := TNdfdLib.Create;
  SetupMenu;
  AddLayers;
  SetupTransformation;
end;


procedure TMainForm.FormDestroy(Sender: TObject);
begin
  Ndfd.Free;
  if Assigned(LatLonTrans) then
    OCTDestroyCoordinateTransformation(LatLonTrans);
end;


procedure TMainForm.AboutCmdClick(Sender: TObject);
begin
  MessageDlg(ProgramName + ', version ' + ProgramVersion + #10#10 +
             'Displays weather forecast for locations on map.'#10#10 +
             'Copyright ' + #$C2#$A9 + ' 2015 Phil Hess',
             mtInformation, [mbOK], 0);
  //TODO: Proper About box.
end;


procedure TMainForm.HelpCmdClick(Sender: TObject);
begin
  MessageDlg('Click the Update Forecast button on the toolbar to update ' +
             'the weather forecast for the map''s labeled locations.'#10#10 +
             'You can also click any point on the map to see the point''s ' +
             'forecast.',
             mtInformation, [mbOK], 0);
  //TODO: Proper help.
end;


procedure TMainForm.UpdateBtnClick(Sender: TObject);
var
  FeatDef   : OGRFeatureDefnH;
  LblFldIdx : Integer;
  Fmt       : TFormatSettings;
  TodayStr  : string;
  Feat      : OGRFeatureH;
  FeatGeom  : OGRGeometryH;
  GeoX      : Double;
  GeoY      : Double;
  GeoZ      : Double;
begin
  if not Assigned(LatLonTrans) then
    Exit;

  FeatDef := OGR_L_GetLayerDefn(LocLayer.LayerH);
  LblFldIdx := OGR_FD_GetFieldIndex(
                FeatDef, PAnsiChar(AnsiString(LocLayer.LabelField)));

  Fmt.DecimalSeparator := '.';
  TodayStr := FormatDateTime('mmm d', Today);

  try
    Screen.Cursor := crHourGlass;
    OGR_L_ResetReading(LocLayer.LayerH);
    repeat
      Feat := OGR_L_GetNextFeature(LocLayer.LayerH);
      if Assigned(Feat) then
        begin
        FeatGeom := OGR_F_GetGeometryRef(Feat);
        if Assigned(FeatGeom) then
          begin
          OGR_G_GetPoint(FeatGeom, 0, @GeoX, @GeoY, @GeoZ);
          OCTTransform(LatLonTrans, 1, @GeoX, @GeoY, @GeoZ);
          if Ndfd.LoadForecast(FloatToStr(GeoY, Fmt), FloatToStr(GeoX, Fmt), 
                               False) then
            begin
            OGR_F_SetFieldString(Feat, LblFldIdx,
                                 PAnsiChar(TodayStr + ': ' +
                                           'High ' + Ndfd.GetMaxTemp(1) +
                                           ', Low ' + Ndfd.GetMinTemp(1)));
            OGR_L_SetFeature(LocLayer.LayerH, Feat);
            end;
          end;
        OGR_F_Destroy(Feat);
        end;
    until not Assigned(Feat);
  finally
    Screen.Cursor := crDefault;
  end;

  Map.Refresh;
end;


procedure TMainForm.MapMouseUp(Sender: TObject; Button: TMouseButton;
                               Shift: TShiftState; X, Y: Integer);
var
  ProjPt : TP4G_Point;
  MapPt  : TPoint;
  OutX   : Double;
  OutY   : Double; 
  OutZ   : Double;
  Fmt    : TFormatSettings;
begin
  ProjPt := Map.MapToProjCoords(X, Y);
//  DebugLn('Clicked ' + IntToStr(X) + ', ' + IntToStr(Y) + ' = ' +
//          DbgS(ProjPt.X) + ', ' + DbgS(ProjPt.Y));
  MapPt := Map.ProjToMapCoords(ProjPt.X, ProjPt.Y);
//  DebugLn('Roundtripped ' + IntToStr(MapPt.X) + ', ' + IntToStr(MapPt.Y));

  if not Assigned(LatLonTrans) then
    Exit;

  OutX := ProjPt.X;
  OutY := ProjPt.Y;
  OutZ := 0;
  OCTTransform(LatLonTrans, 1, @OutX, @OutY, @OutZ);
//  DebugLn('Lat/Lon = ' + DbgS(OutY) + ', ' + DbgS(OutX));

  Fmt.DecimalSeparator := '.';
  if Ndfd.LoadForecast(FloatToStr(OutY, Fmt), FloatToStr(OutX, Fmt), False) then
    begin
//    DebugLn('High/Low = ' + Ndfd.GetMaxTemp(1) + '/' + Ndfd.GetMinTemp(1));
    MessageDlg('You clicked latitude ' + FloatToStrF(OutY, ffFixed, 20, 3, Fmt) + 
               ', longitude ' + FloatToStrF(OutX, ffFixed, 20, 3, Fmt) + #10#10 +
               'Forecast high is ' + Ndfd.GetMaxTemp(1) + ', low is ' +
               Ndfd.GetMinTemp(1),
               mtInformation, [mbOK], 0);
    end
  else
    MessageDlg('Unable to retrieve forecast.'#10#10 + Ndfd.GetLastError,
               mtError, [mbOK], 0);
end;


procedure TMainForm.SetupMenu;
begin
  MainMenu := TMainMenu.Create(Self);
  Menu := MainMenu;

{$IFDEF DARWIN}
  AppMenu := TMenuItem.Create(Self);  {Application menu}
  AppMenu.Caption := #$EF#$A3#$BF;  {Unicode Apple logo char}
  MainMenu.Items.Insert(0, AppMenu);

  AppAboutCmd := TMenuItem.Create(Self);
  AppAboutCmd.Caption := 'About ' + ProgramName;
  AppAboutCmd.OnClick := {$IFDEF FPC_OBJFPC}@{$ENDIF}AboutCmdClick;
  AppMenu.Add(AppAboutCmd);  {Add About as item in application menu}
{$ENDIF}

  HelpMenu := TMenuItem.Create(Self);
  HelpMenu.Caption := 'Help';
{$IFNDEF DARWIN}
  MainMenu.Items.Insert(0, HelpMenu);
{$ELSE}
  MainMenu.Items.Insert(1, HelpMenu);
{$ENDIF}
  HelpHelpCmd := TMenuItem.Create(Self);
  HelpHelpCmd.Caption := ProgramName + ' Help';
  HelpHelpCmd.OnClick := {$IFDEF FPC_OBJFPC}@{$ENDIF}HelpCmdClick;
  HelpMenu.Add(HelpHelpCmd);

{$IFNDEF DARWIN}
  HelpAboutCmd := TMenuItem.Create(Self);
  HelpAboutCmd.Caption := 'About ' + ProgramName;
  HelpAboutCmd.OnClick := {$IFDEF FPC_OBJFPC}@{$ENDIF}AboutCmdClick;
  HelpMenu.Add(HelpAboutCmd);
{$ENDIF}
end;


procedure TMainForm.AddLayers;
var
  LayerPath  : string;
  SoilLayer  : TP4G_VectorLayer;
  TopoLayer  : TP4G_ImageLayer;
  PhotoLayer : TP4G_ImageLayer;
begin
  LayerPath := ExtractFilePath(ParamStr(0));
{$IFDEF DARWIN}
  LayerPath := LayerPath + '../Resources/';
  //TODO: get path to app bundle's Resources folder properly
  // (eg, use NSBundle.mainBundle.resourcePath)
{$ENDIF}

  Map.Legend := Legend;

   //add layers in top to bottom order
  LocLayer := Map.AddVectorLayer(LayerPath + 'locations.shp');
  if Assigned(LocLayer) then
    begin
    LocLayer.PenColor := clMaroon;
    LocLayer.PenWidth := 2;
    LocLayer.LabelField := 'FORECAST';
    LocLayer.LabelFont.Color := clMaroon;
    LocLayer.LabelFont.Height := -15;
    end;

  SoilLayer := Map.AddVectorLayer(LayerPath + 'soils.shp');
  if Assigned(SoilLayer) then
    begin
    SoilLayer.PenColor := clYellow;
    SoilLayer.PenWidth := 2;
    SoilLayer.LabelField := 'musym';
    SoilLayer.LabelFont.Color := clYellow;
    SoilLayer.Visible := False;
    end;

  TopoLayer := Map.AddImageLayer(LayerPath + 'topo.jpg');
  if Assigned(TopoLayer) then
    TopoLayer.Visible := False;

  PhotoLayer := Map.AddImageLayer(LayerPath + 'photo.jpg');
end;


procedure TMainForm.SetupTransformation;
var
  LatLonSrs : OGRSpatialReferenceH;
begin
  if not Assigned(LocLayer) then
    Exit;
  LatLonSrs := OSRNewSpatialReference('');
  OSRSetWellKnownGeogCS(LatLonSrs, 'WGS84');
  LatLonTrans := OCTNewCoordinateTransformation(
                  OGR_L_GetSpatialRef(LocLayer.LayerH), LatLonSrs);
  OSRRelease(LatLonSrs);
end;


end.
