program testshapes;

 {
   Create a point shapefile from array of lat/lon coordinates
    and populate the shapefile's attributes with each point's 
    location name and weather forecast.
   Optionally, transform points to a different geographic
    coordinate system or projection.
  }

{$IFDEF FPC}
 {$mode objfpc}{$H+}
{$ELSE}
 {$APPTYPE CONSOLE}
{$ENDIF}

uses
  SysUtils,
  DateUtils,
  NdfdLib,
  GDAL;

type
  TLocRec = record
    Loc : string;
    Lat : string;
    Lon : string;
    end;

const
  MaxLocs = 4;
  Locations : array [1..MaxLocs] of TLocRec =
    ((Loc:'Ouray';              Lat:'38.023'; Lon:'-107.671'),
     (Loc:'Silverton';          Lat:'37.812'; Lon:'-107.665'),
     (Loc:'Kendall Mtn summit'; Lat:'37.794'; Lon:'-107.642'),
     (Loc:'Telluride';          Lat:'37.937'; Lon:'-107.811'));
   // Note how forecast for the Kendall Mountain summit is about the same as
   //  the nearby town of Silverton, even though the summit is 1000m higher.
   //  Does the NDFD weather forecast model ignore elevation differences?

var
  Driver   : OGRSFDriverH;
  DataSrc  : OGRDataSourceH;
  InSrs    : OGRSpatialReferenceH;
  OutSrs   : OGRSpatialReferenceH;
  Trans    : OGRCoordinateTransformationH;
  Layer    : OGRLayerH;
  LocFld   : OGRFieldDefnH;
  ForeFld  : OGRFieldDefnH;
  TodayStr : string;
  Ndfd     : TNdfdLib;
  LocIdx   : Integer;
  Feat     : OGRFeatureH;
  Point    : OGRGeometryH;
  x, y, z  : Double;

begin
   //Note this is GDAL 1.x-specific code. http://gdal.org/1.11/index.html

  OGRRegisterAll();

  DataSrc := nil;
  OutSrs := nil;
  Trans := nil;
  Ndfd := nil;

  try
    Driver := OGRGetDriverByName('ESRI Shapefile');
    if not Assigned(Driver) then
      begin
      WriteLn('Driver not available');
      Exit;
      end;

    DeleteFile('locations.shp');
    DataSrc := OGR_Dr_CreateDataSource(Driver, 'locations.shp', nil);
    if not Assigned(DataSrc) then
      begin
      WriteLn('Error creating data source');
      Exit;
      end;

    InSrs := OSRNewSpatialReference('');
    OSRSetWellKnownGeogCS(InSrs, 'WGS84');

    OutSrs := OSRNewSpatialReference('');

     // Note: GDAL may not be able to find its gcs.csv, etc. files
     //  (eg, if not using GDAL.framework). Can set path to gcs.csv 
     //  with GDAL_DATA environment variable.

(*
     // GDAL can't create 3857 .prj file like this.
     //  "ERROR 6: No translation for WGS 84 / Pseudo-Mercator to PROJ.4 format is known."
     //  See http://trac.osgeo.org/gdal/ticket/3962
    OSRSetProjCS(OutSrs, 'WGS_1984_Web_Mercator_Auxiliary_Sphere');
    OSRSetWellKnownGeogCS(OutSrs, 'EPSG:3857');
    OSRSetProjection(OutSrs, 'WGS 84 / Pseudo-Mercator');

     // This appears to work but resulting points aren't very accurate.
     //  See this FAQ:
     //  http://trac.osgeo.org/proj/wiki/FAQ#ChangingEllipsoidWhycantIconvertfromWGS84toGoogleEarthVirtualGlobeMercator
//    OSRImportFromEPSG(OutSrs, 3857);
*)

     // Converting to UTM projection works and is accurate:
    OSRSetProjCS(OutSrs, 'UTM 13 (NAD83) in northern hemisphere.');
    OSRSetWellKnownGeogCS(OutSrs, 'NAD83');
    OSRSetUTM(OutSrs, 13, 1);

    Trans := OCTNewCoordinateTransformation(InSrs, OutSrs);
    if not Assigned(Trans) then
      begin
      OSRRelease(OutSrs);
      OutSrs := InSrs;
      end
    else
      OSRRelease(InSrs);

    Layer := OGR_DS_CreateLayer(DataSrc, 'locations', OutSrs, wkbPoint, nil);
    if not Assigned(Layer) then
      begin
      WriteLn('Error creating layer');
      Exit;
      end;

    LocFld := OGR_Fld_Create('LOCATION', OFTString);
    try
      OGR_Fld_SetWidth(LocFld, 30);
      OGR_L_CreateField(Layer, LocFld, 1);
    finally
      OGR_Fld_Destroy(LocFld);
    end;

    ForeFld := OGR_Fld_Create('FORECAST', OFTString);
    try
      OGR_Fld_SetWidth(ForeFld, 30);
      OGR_L_CreateField(Layer, ForeFld, 1);
    finally
      OGR_Fld_Destroy(ForeFld);
    end;

    Ndfd := TNdfdLib.Create;

    TodayStr := FormatDateTime('mmm d', Today);

    for LocIdx := 1 to MaxLocs do
      begin
      Feat := OGR_F_Create(OGR_L_GetLayerDefn(Layer));
      try
        OGR_F_SetFieldString(Feat, OGR_F_GetFieldIndex(Feat, 'LOCATION'),
                             PAnsiChar(Locations[LocIdx].Loc));

        if Ndfd.LoadForecast(Locations[LocIdx].Lat, Locations[LocIdx].Lon, False) then
          begin
          OGR_F_SetFieldString(Feat, OGR_F_GetFieldIndex(Feat, 'FORECAST'),
                               PAnsiChar(TodayStr + ': ' +
                                         'High ' + Ndfd.GetMaxTemp(1) +
                                         ', Low ' + Ndfd.GetMinTemp(1)));
          end;

        x := StrToFloat(Locations[LocIdx].Lon);
        y := StrToFloat(Locations[LocIdx].Lat);
        z := 0;
        if Assigned(Trans) then
          OCTTransform(Trans, 1, @x, @y, @z)
        else
          WriteLn('Skipping transformation');

        Point := OGR_G_CreateGeometry(wkbPoint);
        try
          OGR_G_SetPoint_2D(Point, 0, x, y);
          OGR_F_SetGeometry(Feat, Point); 
        finally
          OGR_G_DestroyGeometry(Point);
        end;

        OGR_L_CreateFeature(Layer, Feat);
      finally
        OGR_F_Destroy(Feat);
        end;
      end;

  finally
    Ndfd.Free;
    OCTDestroyCoordinateTransformation(Trans);
    OSRRelease(OutSrs);
    OGR_DS_Destroy(DataSrc);
    OGRCleanupAll;
  end;

end.
