unit NdfdForecast;

{
  Class for working with NDFD Web service and weather forecast data.

  Author:    Phil Hess.
  Copyright: Copyright (C) Phil Hess.
  License:   Modified LGPL (see Free Pascal's rtl/COPYING.FPC).

  NDFD reference: http://www.usgovxml.com/DataService.aspx?ds=NDFD
}

{$IFDEF FPC}
  {$mode objfpc}{$H+}
{$ENDIF}

interface

uses
  SysUtils,
  DateUtils,
{$IFDEF UseNdfdSoap}
  NdfdSoap;
{$ELSE}
  NdfdRest;
{$ENDIF}

const
  MaxNdfdFcstDays = 5;

type
  TNdfdData = record
    MaxTemp : array [1..MaxNdfdFcstDays] of string;  //use strings since values
    MinTemp : array [1..MaxNdfdFcstDays] of string;  // could be missing
    end;

  TNdfdForecast = class
  private
    FDataFmt  : TFormatSettings;
    FForecast : TNdfdData;
    FErrorMsg : string;
    function GetMaxTemp(DayNum : Integer) : string;
    function GetMinTemp(DayNum : Integer) : string;
  public
    constructor Create;
    function LoadForecast(const Latitude  : string;
                          const Longitude : string;
                                GetMetric : Boolean) : Boolean;
    property MaxTemp[DayNum : Integer] : string read GetMaxTemp;
    property MinTemp[DayNum : Integer] : string read GetMinTemp;
    property ErrorMsg : string read FErrorMsg;
  end;


implementation

constructor TNdfdForecast.Create;
begin
  inherited;
  FDataFmt.DecimalSeparator := '.';  //for StrToFloat conversions
end;

function TNdfdForecast.GetMaxTemp(DayNum : Integer) : string;
begin
  Result := '';
  if (DayNum > 0) and (DayNum <= MaxNdfdFcstDays) then  {Within forecast range?}
    Result := FForecast.MaxTemp[DayNum];
end;

function TNdfdForecast.GetMinTemp(DayNum : Integer) : string;
begin
  Result := '';
  if (DayNum > 0) and (DayNum <= MaxNdfdFcstDays) then  {Within forecast range?}
    Result := FForecast.MinTemp[DayNum];
end;


function TNdfdForecast.LoadForecast(const Latitude  : string;
                                    const Longitude : string;
                                          GetMetric : Boolean) : Boolean;
 {Load forecast for specified latitude/longitude and units.
  If successful, return True to indicate forecast data can be 
   retrieved via MaxTemp/MinTemp properties; if failure, return 
   False to indicate error message can be retrieved via ErrorMsg 
   property.

  Note using strings rather than Double for lat/lon, meaning we have
   possibility of supporting deg-min-sec values. But strings mean have 
   to worry about decimal separator char, which is set in constructor.}
var
  LatDbl   : Double;
  LonDbl   : Double;
  DayNum   : Integer;
  FcstXml  : string;
  LoopCnt  : Integer;
  SrchTag  : string;
  SrchPos  : Integer;
  ValCnt   : Integer;
  ValsDone : Boolean;
  ValStr   : string;
begin
  Result := False;

  if not TryStrToFloat(Latitude, LatDbl, FDataFmt) then
    begin
    FErrorMsg := 'Latitude is invalid number.';
    Exit;
    end;
  if not TryStrToFloat(Longitude, LonDbl, FDataFmt) then
    begin
    FErrorMsg := 'Longitude is invalid number.';
    Exit;
    end;
   //also check if valid for U.S.? maybe Web service does that

   {Blank any previously retrieved data and error message}
  for DayNum := 1 to MaxNdfdFcstDays do
    begin
    FForeCast.MaxTemp[DayNum] := '';
    FForeCast.MinTemp[DayNum] := '';
    end;
  FErrorMsg := '';

  if not GetNdfdForecast(Latitude, Longitude, FormatDateTime('yyyy-mm-dd', Today),
                         MaxNdfdFcstDays, GetMetric, FcstXml, FErrorMsg) then
    Exit;

   {For now, to avoid writing XML parse code twice (once for FPC's
     XMLRead/DOM units and once for Delphi's XMLIntf/XMLDoc units),
     just scan the returned XML string looking for the data we want.}
   //Assumptions:
   // - Maximum temps come first in XML.
   // - XML made up of lines terminated by linefeed (ASCII 10).
   // - Each line of forecast data is of the form: <value>xxx</value>
  for LoopCnt := 1 to 2 do
    begin
    if LoopCnt = 1 then
      SrchTag := '<name>Daily Maximum Temperature</name>'
    else
      SrchTag := '<name>Daily Minimum Temperature</name>';
    SrchPos := Pos(SrchTag, FcstXml);
    if SrchPos = 0 then
      begin
      FErrorMsg := 'Unable to find ' + SrchTag + ' tag.';
      Exit;
      end;
    FcstXml := Trim(Copy(FcstXml, SrchPos + Length(SrchTag), MaxInt));
    ValCnt := 0;
    ValsDone := False;
    repeat
      SrchPos := Pos(#10, FcstXml);  {Look for end of line}
      if SrchPos = 0 then
        SrchPos := Length(FcstXml);
      ValStr := Copy(FcstXml, 1, SrchPos);  {Get line}
      FcstXml := Trim(Copy(FcstXml, SrchPos+1, MaxInt)); {Delete line from XML}
      if Copy(ValStr, 1, 7) <> '<value>' then
        ValsDone := True
      else
        begin
        ValStr := Copy(ValStr, 8, MaxInt);
        SrchPos := Pos('</value>', ValStr);
        if SrchPos = 0 then
          SrchPos := Length(ValStr) + 1;
        ValStr := Trim(Copy(ValStr, 1, SrchPos-1));
        Inc(ValCnt);
        if ValCnt <= MaxNdfdFcstDays then
          begin
          if LoopCnt = 1 then
            FForecast.MaxTemp[ValCnt] := ValStr
          else
            FForecast.MinTemp[ValCnt] := ValStr;
          end;
           //for now, no proofing, rounding or reformatting, 
           // just save what Web service returned
        end;
    until ValsDone;
    end;

  Result := True;
end;


end.
