unit NdfdRest;

{

  Function that retrieves NDFD weather forecast data via REST service.

  https://graphical.weather.gov/xml/rest.php

}

{$IFDEF FPC}
  {$mode objfpc}{$H+}
{$ENDIF}

interface

uses
  SysUtils,
{$IF DEFINED(UseNSHttpClient)}
  ns_url_request;
{$ELSEIF DEFINED(UseSynapseHttpClient)}
  httpsend,
  ssl_openssl;
{$ELSEIF DEFINED(UseIndyHttpClient)}
  IdHTTP;
{$ELSE}  //use FPC's HTTP client as default
  fphttpclient;
{$IFEND}

function GetNdfdForecast(const Latitude  : string;
                         const Longitude : string;
                         const StartDate : string;
                         const NumDays   : Integer;
                         const GetMetric : Boolean;
                           out FcstXml   : string;
                           out ErrorMsg  : string) : Boolean;


implementation

function GetNdfdForecast(const Latitude  : string;
                         const Longitude : string;
                         const StartDate : string;
                         const NumDays   : Integer;
                         const GetMetric : Boolean;
                           out FcstXml   : string;
                           out ErrorMsg  : string) : Boolean;
 {Latitude and longitude are in decimal degrees.
  StartDate is in format yyyy-mm-dd or pass a blank string
   to start with today.
  NumDays is the number of forecast days to request.
  If GetMetric is True, return temperatures in Centigrade;
   if GetMetric is False, return temperatures in Fahrenheit.
  The XML returned by the service is returned in FcstXml.
  If successful, return True as function result; if error, 
   return error in ErrorMsg and False as function result.}

const
  NdfdBaseUrl = 'https://graphical.weather.gov/xml/sample_products/browser_interface/ndfdBrowserClientByDay.php?';

var
  NdfdUrl    : string;

{$IF DEFINED(UseNSHttpClient)}
  HttpClient : TNSHTTPSendAndReceive;
{$ELSEIF DEFINED(UseSynapseHttpClient)}
  HttpClient : THTTPSend;
{$ELSEIF DEFINED(UseIndyHttpClient)}
  HttpClient : TIdHTTP;
{$ELSE}
  HttpClient : TFPHTTPClient;
{$IFEND} 

begin
  Result := False;

   {Single Point Summarized Data}
  NdfdUrl := NdfdBaseUrl + 'lat=' + Latitude +
                           '&lon=' + Longitude;
  if StartDate <> '' then
    NdfdUrl := NdfdUrl + '&startDate=' + StartDate;
  NdfdUrl := NdfdUrl + '&numDays=' + IntToStr(NumDays) +
                       '&Unit=';
  if GetMetric then
    NdfdUrl := NdfdUrl + 'm'
  else
    NdfdUrl := NdfdUrl + 'e';
  NdfdUrl := NdfdUrl + '&format=24+hourly';

{$IF DEFINED(UseNSHttpClient)}
  HttpClient := TNSHTTPSendAndReceive.Create;
  try
    HttpClient.Address := NdfdUrl;
    if not HttpClient.SendAndReceive(FcstXml) then
      begin
      ErrorMsg := HttpClient.LastErrMsg;
      Exit;
      end;  
  finally
    HttpClient.Free;
  end;

{$ELSEIF DEFINED(UseSynapseHttpClient)}
  HttpClient := THTTPSend.Create;
  try
    if HttpClient.HTTPMethod('GET', NdfdUrl) then
      begin
      SetLength(FcstXml, HttpClient.Document.Size);
      if HttpClient.Document.Size > 0 then
        HttpClient.Document.Read(FcstXml[1], HttpClient.Document.Size);
      end
    else
      begin
      ErrorMsg := IntToStr(HttpClient.ResultCode) + ' ' + HttpClient.ResultString;
      Exit;
      end;
  finally
    HttpClient.Free;
  end;

{$ELSEIF DEFINED(UseIndyHttpClient)}
  HttpClient := TIdHTTP.Create(nil);
  try
    try
      FcstXml := HttpClient.Get(NdfdUrl);
    on E:Exception do  //EIdOSSLCouldNotLoadSSLLibrary, EIdHTTPProtocolException, etc.
      begin
      ErrorMsg := E.Message;
      Exit; 
      end;
    end;
  finally
    HttpClient.Free;
  end;

{$ELSE}  {FPC HTTP client}
  HttpClient := TFPHTTPClient.Create(nil);
  try
    try
      FcstXml := HttpClient.Get(NdfdUrl);
    except
      on E:Exception do  //EInOutError, EHTTPClient, etc.
        begin
        ErrorMsg := E.Message;
        Exit; 
        end;
      end;
  finally
    HttpClient.Free;
  end;  
{$IFEND}

  Result := True;
end;


end.
