library ndfdmod;

{
  Python extension module that wraps ndfd library.

  Author:    Phil Hess.
  Copyright: Copyright (C) Phil Hess.
  License:   Modified LGPL (see Free Pascal's rtl/COPYING.FPC).
}

{$IFDEF FPC}
  {$mode objfpc}{$H+}
{$ENDIF}

uses
  SysUtils,
  ndfd_types,
  NdfdLib,
  PyAPI;

{$DEFINE IS32BIT}
{$IFDEF CPUX64}  {Delphi}
  {$UNDEF IS32BIT}
{$ENDIF}
{$IFDEF CPU64}  {FPC}
  {$UNDEF IS32BIT}
{$ENDIF}

{$DEFINE UNIV_OR_32BIT}
{$IFNDEF IS32BIT}
  {$IFNDEF DARWIN}  {Assume OS X library is universal 32/64-bit}
    {$UNDEF UNIV_OR_32BIT}
  {$ENDIF}
{$ENDIF}

const
{$IFDEF UNIV_OR_32BIT}
  {$IFDEF USE_PYTHON3}
  NdfdModName = 'ndfdmod_3';
  {$ELSE}
  NdfdModName = 'ndfdmod';
  {$ENDIF}
{$ELSE}
  {$IFDEF USE_PYTHON3}
  NdfdModName = 'ndfdmod64_3';
  {$ELSE}
  NdfdModName = 'ndfdmod64';
  {$ENDIF}
{$ENDIF}

{$IFDEF IS32BIT}
  kchar = 'k';
{$ELSE}
  kchar = 'K';
{$ENDIF}


function GetLibVersion(Self : PyObject) : PyObject; cdecl;
var
  Buffer : array [0..20] of NdfdChar;
begin
  NdfdGetLibVersion(Buffer, SizeOf(Buffer));
  Result := PyString_FromString(Buffer);
end;

function Init(Self : PyObject;
              Args : PyObject) : PyObject; cdecl;
begin
  Result := PyLong_FromVoidPtr(NdfdInit);
end;

function Uninit(Self : PyObject;
                Args : PyObject) : PyObject; cdecl;
var
  NdfdObj : NdfdPObj;
begin
  if PyArg_ParseTuple(Args, kchar, @NdfdObj) = 0 then
    begin
    Result := nil;
    Exit;
    end;
  Result := PyInt_FromLong(NdfdInt(NdfdUninit(NdfdObj)));
end;

function GetLastError(Self : PyObject;
                      Args : PyObject) : PyObject; cdecl;
var
  NdfdObj : NdfdPObj;
  Buffer  : array [0..200] of NdfdChar;
begin
  if PyArg_ParseTuple(Args, kchar, @NdfdObj) = 0 then
    begin
    Result := nil;
    Exit;
    end;
  NdfdGetLastError(NdfdObj, Buffer, SizeOf(Buffer));
  Result := PyString_FromString(Buffer);
end;

function LoadForecast(Self : PyObject;
                      Args : PyObject) : PyObject; cdecl;
var
  NdfdObj   : NdfdPObj;
  Latitude  : NdfdPChar;
  Longitude : NdfdPChar;
  GetMetric : NdfdBool;
begin
  if PyArg_ParseTuple(Args, kchar + 'ssi', @NdfdObj, @Latitude, @Longitude,
                      @GetMetric) = 0 then
    begin
    Result := nil;
    Exit;
    end;
  Result := PyInt_FromLong(NdfdInt(NdfdLoadForecast(NdfdObj, Latitude,
                                                    Longitude, GetMetric)));
end;

function GetMaxTemp(Self : PyObject;
                    Args : PyObject) : PyObject; cdecl;
var
  NdfdObj     : NdfdPObj;
  DayNum      : NdfdUInt;
  TempBuf     : NdfdPChar;
  TempBufSize : NdfdUInt;
  ValueOkay   : NdfdBool;
begin
  if PyArg_ParseTuple(Args, kchar + 'iz#', @NdfdObj, @DayNum, @TempBuf,
                      @TempBufSize) = 0 then
    begin
    Result := nil;
    Exit;
    end;
  ValueOkay := NdfdGetMaxTemp(NdfdObj, DayNum, TempBuf, TempBufSize);
  Result := PyInt_FromLong(NdfdInt(ValueOkay));
  if (not ValueOkay) and (TempBufSize > 1) then
    StrPCopy(TempBuf, Copy(AnsiString(TempBuf), 1, TempBufSize-1));  //AnsiString?
end;

function GetMinTemp(Self : PyObject;
                    Args : PyObject) : PyObject; cdecl;
var
  NdfdObj     : NdfdPObj;
  DayNum      : NdfdUInt;
  TempBuf     : NdfdPChar;
  TempBufSize : NdfdUInt;
  ValueOkay   : NdfdBool;
begin
  if PyArg_ParseTuple(Args, kchar + 'iz#', @NdfdObj, @DayNum, @TempBuf,
                      @TempBufSize) = 0 then
    begin
    Result := nil;
    Exit;
    end;
  ValueOkay := NdfdGetMinTemp(NdfdObj, DayNum, TempBuf, TempBufSize);
  Result := PyInt_FromLong(NdfdInt(ValueOkay));
  if (not ValueOkay) and (TempBufSize > 1) then
    StrPCopy(TempBuf, Copy(AnsiString(TempBuf), 1, TempBufSize-1));  //AnsiString?
end;



const
  NumFuncs = 7;

var
  Methods : packed array [0..NumFuncs] of PyMethodDef;

{$IFDEF USE_PYTHON3}
  ModuleDef : PyModuleDef;

  {$IFDEF UNIV_OR_32BIT}
procedure PyInit_ndfdmod_3; cdecl;
  {$ELSE}
procedure PyInit_ndfdmod64_3; cdecl;
  {$ENDIF}

{$ELSE}
  {$IFDEF UNIV_OR_32BIT}
procedure initndfdmod; cdecl;
  {$ELSE}
procedure initndfdmod64; cdecl;
  {$ENDIF}
{$ENDIF}

begin
  Methods[0].name := 'GetLibVersion';
  Methods[0].meth := @GetLibVersion;
  Methods[0].flags := METH_VARARGS;
  Methods[0].doc := '';

  Methods[1].name := 'Init';
  Methods[1].meth := @Init;
  Methods[1].flags := METH_VARARGS;
  Methods[1].doc := '';

  Methods[2].name := 'Uninit';
  Methods[2].meth := @Uninit;
  Methods[2].flags := METH_VARARGS;
  Methods[2].doc := '';

  Methods[3].name := 'GetLastError';
  Methods[3].meth := @GetLastError;
  Methods[3].flags := METH_VARARGS;
  Methods[3].doc := '';

  Methods[4].name := 'LoadForecast';
  Methods[4].meth := @LoadForecast;
  Methods[4].flags := METH_VARARGS;
  Methods[4].doc := '';

  Methods[5].name := 'GetMaxTemp';
  Methods[5].meth := @GetMaxTemp;
  Methods[5].flags := METH_VARARGS;
  Methods[5].doc := '';

  Methods[6].name := 'GetMinTemp';
  Methods[6].meth := @GetMinTemp;
  Methods[6].flags := METH_VARARGS;
  Methods[6].doc := '';

  Methods[NumFuncs].name := nil;
  Methods[NumFuncs].meth := nil;
  Methods[NumFuncs].flags := 0;
  Methods[NumFuncs].doc := nil;

{$IFDEF USE_PYTHON3}
  ModuleDef.m_name := NdfdModName;
  ModuleDef.m_doc := nil;
  ModuleDef.m_size := 0;  //?
  ModuleDef.m_methods := @Methods;
  ModuleDef.m_reload := nil;
  ModuleDef.m_traverse := nil;
  ModuleDef.m_clear := nil;
  ModuleDef.m_free := nil;
  PyModule_Create(ModuleDef);
{$ELSE}
  Py_InitModule(NdfdModName, Methods[0]);
{$ENDIF}
end;

exports
{$IFDEF USE_PYTHON3}
  {$IFDEF UNIV_OR_32BIT}
  PyInit_ndfdmod_3;
  {$ELSE}
  PyInit_ndfdmod64_3;
  {$ENDIF}
{$ELSE}
  {$IFDEF UNIV_OR_32BIT}
  initndfdmod;
  {$ELSE}
  initndfdmod64;
  {$ENDIF}
{$ENDIF}
 
end.
