unit P4G_Layer;

{
  Pascal for GIS (P4G) map layers.

  Author:    Phil Hess.
  Copyright: Copyright (C) Phil Hess.
  License:   Modified LGPL (see Free Pascal's rtl/COPYING.FPC).

}

{$IFDEF FPC}
  {$mode objfpc}{$H+}
{$ENDIF}

interface

uses
  SysUtils,
  Classes,
  Graphics,
  Controls,
  GDAL;  //for now, links Lazarus to GDAL framework/library too

type
  TP4G_Point = record
    X : Double;
    Y : Double;
  end;

  TP4G_Extent = record
    XMin  : Double;
    YMin  : Double;
    XMax  : Double;
    YMax  : Double;
    XSize : Double;
    YSize : Double;
    end;


  TP4G_Layer = class
  private
    FFileName : string;  //what if wanted to use in-memory data? leave blank?
    FExtent   : TP4G_Extent;
    FMapCtrl  : TControl;
    FVisible  : Boolean;
    procedure SetFileName(const AFileName : string); virtual; abstract;
    function GetExtent : TP4G_Extent; virtual; abstract;
    procedure SetVisible(IsVisible : Boolean);
  public
    property FileName : string read FFileName write SetFileName;
    property Extent   : TP4G_Extent read GetExtent;  {Layer's geographic extent}
    property MapCtrl  : TControl read FMapCtrl write FMapCtrl;  {Layer renderer}
    property Visible  : Boolean read FVisible write SetVisible;
  end;


  TP4G_ImageLayer = class(TP4G_Layer)
  private
    FDataset    : GDALDatasetH;
    FXSize      : Integer;
    FYSize      : Integer;
    GeoTrans    : array [0..5] of Double;
    InvGeoTrans : array [0..5] of Double;
    procedure SetFileName(const AFileName : string); override;
    function GetExtent : TP4G_Extent; override;
  public
    constructor Create;
    destructor Destroy; override;
    function RastToProjCoords(X : Integer;
                              Y : Integer) : TP4G_Point;
    function ProjToRastCoords(X : Double;
                              Y : Double) : TPoint;
    property Dataset : GDALDatasetH read FDataset;
    property XSize   : Integer read FXSize;  {Number of pixels in raster line}
    property YSize   : Integer read FYSize;  {Number of raster lines}
  end;


  TP4G_VectorLayer = class(TP4G_Layer)
  private
    FDataSrc    : OGRDataSourceH;
    FLayerH     : OGRLayerH;
    FLabelField : string;
    FLabelFont  : TFont;
    FPenColor   : TColor;
    FPenWidth   : Integer;
    procedure SetFileName(const AFileName : string); override;
    function GetExtent : TP4G_Extent; override;
  public
    constructor Create;
    destructor Destroy; override;
    property DataSrc    : OGRDataSourceH read FDataSrc;
    property LayerH     : OGRLayerH read FLayerH;
    property LabelField : string read FLabelField write FLabelField;
                           {Name of field to use for feature (shape) labels}
    property LabelFont  : TFont read FLabelFont;
    property PenColor   : TColor read FPenColor write FPenColor;
    property PenWidth   : Integer read FPenWidth write FPenWidth;  {Line thickness}
  end;


implementation

procedure TP4G_Layer.SetVisible(IsVisible : Boolean);
begin
  if IsVisible = Visible then
    Exit;
  FVisible := IsVisible;
  if Assigned(MapCtrl) then
    MapCtrl.Refresh;
end;


constructor TP4G_ImageLayer.Create;
begin
end;


destructor TP4G_ImageLayer.Destroy;
begin
  GDALClose(Dataset);
  inherited Destroy;
end;


procedure TP4G_ImageLayer.SetFileName(const AFileName : string);
var
  ProjPt : TP4G_Point;
begin
  if Assigned(Dataset) then
    GDALClose(Dataset);
  FFileName := '';
  FDataset := GDALOpen(PAnsiChar(AnsiString(AFileName)), GA_ReadOnly);  //assume won't be editing
  if not Assigned(Dataset) then
    Exit;

  FFileName := AFileName;
  GDALGetGeoTransform(Dataset, @GeoTrans);  //check if error
  GDALInvGeoTransform(@GeoTrans, @InvGeoTrans);
  FXSize := GDALGetRasterXSize(Dataset);
  FYSize := GDALGetRasterYSize(Dataset);
  ProjPt := RastToProjCoords(0, YSize-1);  {Lower-left corner}
  FExtent.XMin := ProjPt.X;
  FExtent.YMin := ProjPt.Y;
  ProjPt := RastToProjCoords(XSize-1, 0);  {Upper-right corner}
  FExtent.XMax := ProjPt.X;
  FExtent.YMax := ProjPt.Y;
  FExtent.XSize := Extent.XMax - Extent.XMin;
  FExtent.YSize := Extent.YMax - Extent.YMin;
end;


function TP4G_ImageLayer.GetExtent : TP4G_Extent;
begin
  Result := FExtent;
end;


function TP4G_ImageLayer.RastToProjCoords(X : Integer;
                                          Y : Integer) : TP4G_Point;
begin
  GDALApplyGeoTransform(@GeoTrans, X, Y, @Result.X, @Result.Y);
end;


function TP4G_ImageLayer.ProjToRastCoords(X : Double;
                                          Y : Double) : TPoint;
var
  RastX : Double;
  RastY : Double;
begin
  GDALApplyGeoTransform(@InvGeoTrans, X, Y, @RastX, @RastY);
  Result.X := Round(RastX);
  Result.Y := Round(RastY);
end;


constructor TP4G_VectorLayer.Create;
begin
  PenColor := clBlack;
  PenWidth := 1;
  FLabelFont := TFont.Create;  //does this set default font properties?
end;


destructor TP4G_VectorLayer.Destroy;
begin
  FLabelFont.Free;
  OGR_DS_Destroy(DataSrc);
  inherited Destroy;
end;


procedure TP4G_VectorLayer.SetFileName(const AFileName : string);
var
  LayerName : string;
begin
  if Assigned(DataSrc) then
    OGR_DS_Destroy(DataSrc);
  FFileName := '';
  FDataSrc := OGROpen(PAnsiChar(AnsiString(AFileName)), 1, nil);  //open read-write
  if not Assigned(FDataSrc) then
    Exit;

  FFileName := AFileName;
  LayerName :=
   Copy(ExtractFileName(FileName), 1,
   Length(ExtractFileName(FileName))-Length(ExtractFileExt(FileName)));

  FLayerH := OGR_DS_GetLayerByName(DataSrc, PAnsiChar(AnsiString(LayerName)));
end;


function TP4G_VectorLayer.GetExtent : TP4G_Extent;
var
  Env : OGREnvelope;
begin
   {Since might be slow to scan large number of features (shapes),
     only determine extent when requested.}
  if (FExtent.XMax = 0) and (FExtent.XMin = 0) then
    begin
    if OGR_L_GetExtent(LayerH, @Env, 1) <> OGRERR_FAILURE then
      begin
      FExtent.XMin := Env.MinX;
      FExtent.YMin := Env.MinY;
      FExtent.XMax := Env.MaxX;
      FExtent.YMax := Env.MaxY;
//writeln(FExtent.XMin:9:0, ' ', FExtent.YMin:9:0, ' ', FExtent.XMax:9:0, ' ', FExtent.YMax:9:0);
      FExtent.XSize := Env.MaxX - Env.MinX;
      FExtent.YSize := Env.MaxY - Env.MinY;
      end;
    end;
  Result := FExtent;
end;


end.
