unit P4G_Map;

{
  Pascal for GIS (P4G) map control.

  Author:    Phil Hess.
  Copyright: Copyright (C) Phil Hess.
  License:   Modified LGPL (see Free Pascal's rtl/COPYING.FPC).
}

{
  GDAL reference:  http://gdal.org/1.11/index.html
}


{$IFDEF FPC}
  {$mode objfpc}{$H+}
{$ENDIF}

interface

uses
  SysUtils,
  Classes, 
  {$IFDEF LCL}
  LResources,
  LclType,
  LclIntf,
  IntfGraphics,
  {$ENDIF}
  Graphics,
  Controls,
  P4G_Layer,
  P4G_Legend,
  GDAL;  //for now, links Lazarus to GDAL framework/library too

type
  TP4G_CustomMap = class(TCustomControl)
  private
    LayerLst  : TStringList;
    FExtent   : TP4G_Extent;
    MapWidth  : Integer;  {Portion of map control used to draw map}
    MapHeight : Integer;
    FLegend   : TP4G_Legend;
    function GetLayer(LayerIdx : Integer) : TP4G_Layer;
    procedure CalcMapSize;
    procedure SetLegend(ALegend : TP4G_Legend);
    procedure PaintImageLayer(Layer : TP4G_ImageLayer);
    procedure PaintVectorLayer(Layer : TP4G_VectorLayer);
  protected
    procedure Paint; override;
  public
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
    procedure Clear;
    function MapToProjCoords(X : Integer;
                             Y : Integer) : TP4G_Point;
    function ProjToMapCoords(X : Double;
                             Y : Double) : TPoint;
    function AddImageLayer(const FileName : string) : TP4G_ImageLayer;
    function AddVectorLayer(const FileName : string) : TP4G_VectorLayer;
    property Extent : TP4G_Extent read FExtent write FExtent;  {Map's geographic extent}
    property Layers[LayerIdx : Integer] : TP4G_Layer read GetLayer;
    property Legend : TP4G_Legend read FLegend write SetLegend;  {Associated legend control}
  end;


  TP4G_Map = class(TP4G_CustomMap)
  published
    property Align;
    property Anchors;
    property Constraints;
    property OnMouseUp;
  end;


procedure Register;


implementation

constructor TP4G_CustomMap.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);

  ControlStyle := ControlStyle + [csClickEvents, csFramed] - [csCaptureMouse];

  Height := 140;
  Width := 200;
  TabStop := False;

  LayerLst := TStringList.Create;
  FLegend := TP4G_Legend.Create(self);
  FLegend.MapCtrl := self;

  if not (csDesigning in ComponentState) then
    begin
    GDALAllRegister();
    OGRRegisterAll();
    end;
end;


destructor TP4G_CustomMap.Destroy;
var
  LayerIdx : Integer;
begin
  for LayerIdx := 0 to LayerLst.Count-1 do
    Layers[LayerIdx].Free;
  LayerLst.Free;
  OGRCleanupAll;
  inherited Destroy;
end;


procedure TP4G_CustomMap.Clear;
var
  LayerIdx : Integer;
begin
  for LayerIdx := 0 to LayerLst.Count-1 do
    Layers[LayerIdx].Free;
  LayerLst.Clear;
  FillChar(FExtent, SizeOf(FExtent), 0);
  MapWidth := 0;
  MapHeight := 0;
  Legend.Clear;
  Refresh;
end;


function TP4G_CustomMap.GetLayer(LayerIdx : Integer) : TP4G_Layer;
begin
  Result := nil;
  if (LayerIdx < 0) or (LayerIdx >= LayerLst.Count) then
    Exit;
  Result := TP4G_Layer(LayerLst.Objects[LayerIdx]);
end;


procedure TP4G_CustomMap.SetLegend(ALegend : TP4G_Legend);
var
  LayerIdx : Integer;
begin
  if not Assigned(ALegend) then
    Exit;

   {Transfer any items from map's current/default legend}
  for LayerIdx := 0 to Legend.Count-1 do
    begin
    ALegend.AddItem(Legend.Items[LayerIdx],
                    Legend.Items.Objects[LayerIdx]);
    ALegend.Checked[LayerIdx] := Legend.Checked[LayerIdx];
    end;
  ALegend.MapCtrl := self;
  Legend.Free;
  FLegend := ALegend;
end;


procedure TP4G_CustomMap.CalcMapSize;
//var
//  NewExtent : TP4G_Extent;
begin
  MapWidth := Width;
  MapHeight := Height;
   {Since map extent may not have same width:height ratio as map control,
     determine width or height of section of control that will actually be painted}
  if Width / Height > Extent.XSize / Extent.YSize then
    MapWidth := Trunc(Width * (Extent.XSize / Extent.YSize) / (Width / Height))
  else
    MapHeight := Trunc(Height * (Width / Height) / (Extent.XSize / Extent.YSize));
(*
  NewExtent := Extent;
  if Width / Height > Extent.XSize / Extent.YSize then
    begin
    NewExtent.XSize := Extent.XSize * (Width / Height) / (Extent.XSize / Extent.YSize);
    NewExtent.XMax := NewExtent.XMin + NewExtent.XSize;
    if NewExtent.XSize > Extent.XSize then
      Extent := NewExtent;
    end
  else
    begin
    NewExtent.YSize := Extent.YSize * (Extent.XSize / Extent.YSize) / (Width / Height);
    NewExtent.YMin := NewExtent.YMax - NewExtent.YSize;
    if NewExtent.YSize > Extent.YSize then
      Extent := NewExtent;
    end;
*)
end;


function TP4G_CustomMap.MapToProjCoords(X : Integer;
                                        Y : Integer) : TP4G_Point;
begin
  Result.X := Extent.XMin + Extent.XSize * (X + 0.5) / MapWidth;
  Result.Y := Extent.YMin + Extent.YSize * (MapHeight - Y - 0.5) / MapHeight;
end;


function TP4G_CustomMap.ProjToMapCoords(X : Double;
                                        Y : Double) : TPoint;
begin
  Result.X := Trunc((X - Extent.XMin) / Extent.XSize * MapWidth);
  Result.Y := Trunc(MapHeight - (Y - Extent.YMin) / Extent.YSize * MapHeight);
   //Minimum map (and raster) Y is maximum geo coord Y (latitude increases S->N).
end;


function TP4G_CustomMap.AddImageLayer(const FileName : string) : TP4G_ImageLayer;
 {Create an image layer and add it to map so it's on top of other layers.}
  //return err msg? use CPLError?
var
  Layer     : TP4G_ImageLayer;
  LayerName : string;
begin
  Result := nil;
  if not FileExists(FileName) then
    Exit;

  Layer := TP4G_ImageLayer.Create;
  Layer.FileName := FileName;
  if not Assigned(Layer.Dataset) then
    begin
    Layer.Free;
    Exit;
    end;

  if (Extent.XMax = 0) and (Extent.XMin = 0) then
    begin  {Set map's initial extent based on first image layer added}
    FExtent := Layer.Extent;
//writeln(Extent.XMin:9:0, ' ', Extent.YMin:9:0, ' ', Extent.XMax:9:0, ' ', Extent.YMax:9:0);
    CalcMapSize;
    end;

  LayerName := 
   Copy(ExtractFileName(FileName), 1,
        Length(ExtractFileName(FileName))-Length(ExtractFileExt(FileName)));

  LayerLst.AddObject(LayerName, Layer);
  Legend.Items.AddObject(LayerName, Layer);
  Legend.Checked[Legend.Items.Count-1] := True;
  Layer.Visible := True;
  Layer.MapCtrl := self;  //set after Visible to avoid triggering Paint

  Result := Layer;
end;


function TP4G_CustomMap.AddVectorLayer(const FileName : string) : TP4G_VectorLayer;
 {Create a vector layer and add it to map so it's on top of other layers.}
  //return err msg? use CPLError?
var
  Layer     : TP4G_VectorLayer;
  LayerName : string;
begin
  Result := nil;
  if not FileExists(FileName) then
    Exit;

  Layer := TP4G_VectorLayer.Create;
  Layer.FileName := FileName;
  if not Assigned(Layer.DataSrc) then
    begin
    Layer.Free;
    Exit;
    end;

  LayerName := 
   Copy(ExtractFileName(FileName), 1,
        Length(ExtractFileName(FileName))-Length(ExtractFileExt(FileName)));

  LayerLst.AddObject(LayerName, Layer);
  Legend.Items.AddObject(LayerName, Layer);
  Legend.Checked[Legend.Items.Count-1] := True;
  Layer.Visible := True;
  Layer.MapCtrl := self;  //set after Visible to avoid triggering Paint

  Result := Layer;
end;


procedure TP4G_CustomMap.PaintImageLayer(Layer : TP4G_ImageLayer);
type
  TRastArray  = array[0..65535] of Byte;  //what's the maximum to use here?
var
  RBand       : GDALRasterBandH;
  GBand       : GDALRasterBandH;
  BBand       : GDALRasterBandH;
  BandNum     : Integer;
  ColorInterp : Integer;
  RastHeight  : Integer;
  RastWidth   : Integer;
  RLine       : ^TRastArray;
  GLine       : ^TRastArray;
  BLine       : ^TRastArray;
  Bitmap      : TBitmap;
  LazImg      : TLazIntfImage;  //Note Lazarus-specific class
  LineInc     : Integer;
//  RastLine    : Integer;
  MapPt       : TPoint;
  MapRow      : Integer;
  ProjPt      : TP4G_Point;
  RastPt      : TPoint;
  RastPt2     : TPoint;
  PixToRead   : Integer;
  LinesToRead : Integer;
  MapCol      : Integer;
  ColLeft     : Integer;
  ColRight    : Integer;
  ColsToFill  : Integer;
begin
  RBand := nil;
  GBand := nil;
  BBand := nil;  
  for BandNum := 1 to GDALGetRasterCount(Layer.Dataset) do
    begin
    ColorInterp :=
     GDALGetRasterColorInterpretation(GDALGetRasterBand(Layer.Dataset, BandNum));
    if ColorInterp = GCI_RedBand then
      RBand := GDALGetRasterBand(Layer.Dataset, BandNum)
    else if ColorInterp = GCI_GreenBand then
      GBand := GDALGetRasterBand(Layer.Dataset, BandNum)
    else if ColorInterp = GCI_BlueBand then
      BBand := GDALGetRasterBand(Layer.Dataset, BandNum);
    end;
  if (not Assigned(RBand)) or (not Assigned(GBand)) or
     (not Assigned(BBand)) then
    Exit;  //for now only support image files that have RGB bands
  if GDALGetRasterDataType(RBand) <> GDT_Byte then
    Exit;  //for now only support image files with band values of type byte

  RastHeight := Layer.YSize;
  RastWidth := Layer.XSize;

  GetMem(RLine, SizeOf(Byte)*MapWidth);
  GetMem(GLine, SizeOf(Byte)*MapWidth);
  GetMem(BLine, SizeOf(Byte)*MapWidth);

  Bitmap := TBitmap.Create;
  Bitmap.Height := MapHeight;
  Bitmap.Width := MapWidth;
  LazImg := Bitmap.CreateIntfImage;  {Speed up by drawing to bitmap first}

   {If vertical raster resolution is multiple of map resolution, let GDAL do
     "image decimation" of multiple raster lines into a single buffer line.}
  LineInc := RastHeight div MapHeight;
  if LineInc = 0 then
    LineInc := 1;

(*
  try
    RastLine := 0;
    while RastLine < RastHeight do
      begin
      ProjPt := Layer.RastToProjCoords(0, RastLine);
      MapPt := ProjToMapCoords(ProjPt.X, ProjPt.Y);
      if (MapPt.Y >= 0) and (MapPt.Y < MapHeight) then  //safety checks
        begin
         {Let GDAL do "image decimation / replication" of raster area into
           map area buffer (improves both speed and quality of rendering).
           Ie, doesn't matter if RastWidth * LineInc <> MapWidth * 1.}
        GDALRasterIO(RBand, GF_Read, 0, RastLine, RastWidth, LineInc,
                     RLine, MapWidth, 1, GDT_Byte, 0, 0);
        GDALRasterIO(GBand, GF_Read, 0, RastLine, RastWidth, LineInc,
                     GLine, MapWidth, 1, GDT_Byte, 0, 0);
        GDALRasterIO(BBand, GF_Read, 0, RastLine, RastWidth, LineInc,
                     BLine, MapWidth, 1, GDT_Byte, 0, 0);
        for MapCol := 0 to MapWidth-1 do
          begin
          LazImg.Colors[MapCol, MapPt.Y] :=
           TColorToFPColor(
            RGBToColor(RLine^[MapCol], GLine^[MapCol], BLine^[MapCol]));
          end;
        end;
      Inc(RastLine, LineInc);
      if (RastLine + LineInc) >= RastHeight then  {Last map row?}
        LineInc := RastHeight - RastLine;  {Don't read beyond last raster line}
      end;
*)

  try
    MapRow := 0;
    while MapRow < MapHeight do
      begin
      ColLeft := 0;
      ProjPt := MapToProjCoords(ColLeft, MapRow);
      RastPt := Layer.ProjToRastCoords(ProjPt.X, ProjPt.Y);
      if RastPt.X < 0 then
        begin
        RastPt.X := 0;
        ProjPt := Layer.RastToProjCoords(RastPt.X, RastPt.Y);
        MapPt := ProjToMapCoords(ProjPt.X, ProjPt.Y);
        ColLeft := MapPt.X;
        end;

      ColRight := MapWidth-1;
      ProjPt := MapToProjCoords(ColRight, MapRow);
      RastPt2 := Layer.ProjToRastCoords(ProjPt.X, ProjPt.Y);
      if RastPt2.X >= RastWidth then
        begin
        RastPt2.X := RastWidth-1;
        ProjPt := Layer.RastToProjCoords(RastPt2.X, RastPt2.Y);
        MapPt := ProjToMapCoords(ProjPt.X, ProjPt.Y);
        ColRight := MapPt.X;        
        end;

      ColsToFill := ColRight - ColLeft + 1;
      PixToRead := RastPt2.X - RastPt.X + 1;
      LinesToRead := LineInc;
      if (RastPt.Y + LinesToRead) >= RastHeight then
        LinesToRead := RastHeight - RastPt.Y;

      if (RastPt.Y >= 0) and (RastPt.Y < RastHeight) then
        begin
         {Let GDAL do "image decimation / replication" of raster area into
           map area buffer (improves both speed and quality of rendering).
           Ie, doesn't matter if PixToRead * LinesToRead <> ColsToFill * 1.}
        GDALRasterIO(RBand, GF_Read, RastPt.X, RastPt.Y, PixToRead, LinesToRead,
                     @RLine^[ColLeft], ColsToFill, 1, GDT_Byte, 0, 0);
        GDALRasterIO(GBand, GF_Read, RastPt.X, RastPt.Y, PixToRead, LinesToRead,
                     @GLine^[ColLeft], ColsToFill, 1, GDT_Byte, 0, 0);
        GDALRasterIO(BBand, GF_Read, RastPt.X, RastPt.Y, PixToRead, LinesToRead,
                     @BLine^[ColLeft], ColsToFill, 1, GDT_Byte, 0, 0);
        for MapCol := 0 to MapWidth-1 do
          begin
          if (MapCol >= ColLeft) and (MapCol <= ColRight) then
            LazImg.Colors[MapCol, MapRow] :=
             TColorToFPColor(
              RGBToColor(RLine^[MapCol], GLine^[MapCol], BLine^[MapCol]));
          end;
        end;
      Inc(MapRow);
      end;

    Bitmap.LoadFromIntfImage(LazImg);
//    ProjPt := Layer.RastToProjCoords(0, 0);
//    MapPt := ProjToMapCoords(ProjPt.X, ProjPt.Y);
//    Canvas.Draw(MapPt.X, MapPt.Y, Bitmap);
    Canvas.Draw(0, 0, Bitmap);

  finally
    LazImg.Free; 
    Bitmap.Free;
    FreeMem(RLine);
    FreeMem(GLine);
    FreeMem(BLine);
  end;
end;  {PaintImageLayer}


procedure TP4G_CustomMap.PaintVectorLayer(Layer : TP4G_VectorLayer);
var
  FeatDef   : OGRFeatureDefnH;
  LblFldIdx : Integer;
  Feat      : OGRFeatureH;
  FeatGeom  : OGRGeometryH;
  Geom      : OGRGeometryH;
  NumGeoms  : Integer;
  GeomIdx   : Integer;
  PtIdx     : Integer;
  GeoX      : Double;
  GeoY      : Double;
  GeoZ      : Double;
  Pt1       : TPoint;
  Pt2       : TPoint;
  Centroid  : OGRGeometryH;
  LabelStr  : string;
  SaveMode  : Integer;
begin
  Canvas.Pen.Color := Layer.PenColor;
  Canvas.Pen.Width := Layer.PenWidth;

  FeatDef := OGR_L_GetLayerDefn(Layer.LayerH);
  LblFldIdx := OGR_FD_GetFieldIndex(FeatDef, PAnsiChar(AnsiString(Layer.LabelField)));

  OGR_L_ResetReading(Layer.LayerH);
  repeat
    Feat := OGR_L_GetNextFeature(Layer.LayerH);
    if Assigned(Feat) then
      begin
      FeatGeom := OGR_F_GetGeometryRef(Feat);
      if Assigned(FeatGeom) then
        begin
        Geom := FeatGeom;
        NumGeoms := 1;
        if OGR_G_GetGeometryType(FeatGeom) in [wkbPolygon, wkbMultiPolygon] then
          NumGeoms := OGR_G_GetGeometryCount(FeatGeom);
         //possible todo: wkbMultiPoint, wkbMultiLineString, wkbGeometryCollection

        for GeomIdx := 0 to NumGeoms-1 do
          begin
          Centroid := OGR_G_CreateGeometry(wkbPoint);
          OGR_G_Centroid(Geom, Centroid);
          if OGR_G_GetGeometryType(FeatGeom) in [wkbPolygon, wkbMultiPolygon] then
            begin
            Geom := OGR_G_GetGeometryRef(FeatGeom, GeomIdx);
            OGR_G_Centroid(Geom, Centroid);
            if OGR_G_GetGeometryType(FeatGeom) = wkbMultiPolygon then
              Geom := OGR_G_GetGeometryRef(Geom, 0);
            end;

          OGR_G_GetPoint(Geom, 0, @GeoX, @GeoY, @GeoZ);
          Pt1 := ProjToMapCoords(GeoX, GeoY);
          if OGR_G_GetGeometryType(FeatGeom) = wkbPoint then
            Canvas.Pixels[Pt1.X, Pt1.Y] := Layer.PenColor
          else
            begin
            for PtIdx := 1 to OGR_G_GetPointCount(Geom)-1 do
              begin
              OGR_G_GetPoint(Geom, PtIdx, @GeoX, @GeoY, @GeoZ);
              Pt2 := ProjToMapCoords(GeoX, GeoY);
              Canvas.Line(Pt1.X, Pt1.Y, Pt2.X, Pt2.Y);
              Pt1 := Pt2;
              end;
            end;

          if LblFldIdx >= 0 then
            begin
            OGR_G_GetPoint(Centroid, 0, @GeoX, @GeoY, @GeoZ);
            Pt1 := ProjToMapCoords(GeoX, GeoY);
            LabelStr := AnsiString(OGR_F_GetFieldAsString(Feat, LblFldIdx));      
            SaveMode := SetBkMode(Canvas.Handle, TRANSPARENT);
            Canvas.Font := Layer.LabelFont;
            Canvas.TextOut(Pt1.X - Canvas.TextWidth(LabelStr) div 2,
                           Pt1.Y - Canvas.TextHeight(LabelStr) div 2,
                           LabelStr);
            SetBkMode(Canvas.Handle, SaveMode);
            end;
          OGR_G_DestroyGeometry(Centroid);
          end;
        end;
      OGR_F_Destroy(Feat);
      end;
  until not Assigned(Feat);
end;  {PaintVectorLayer}


procedure TP4G_CustomMap.Paint;
var
  LayerIdx : Integer;
  LayerVis : Boolean;
begin
  if csDesigning in ComponentState then
    Exit;
  if LayerLst.Count = 0 then
    Exit;
  if (Extent.XMax = 0) and (Extent.XMin = 0) then
    begin  {Map doesn't yet have an extent, so use first layer's extent}
    FExtent := Layers[0].Extent;
    if (FExtent.XMax = 0) and (FExtent.XMin = 0) then
      Exit;
    end;
  CalcMapSize;
  for LayerIdx := LayerLst.Count-1 downto 0 do
    begin
    LayerVis := Layers[LayerIdx].Visible;
    Legend.Checked[LayerIdx] := LayerVis;  //make sure legend is in synch
    if LayerVis then
      begin
      if Layers[LayerIdx] is TP4G_ImageLayer then
        PaintImageLayer(TP4G_ImageLayer(Layers[LayerIdx]))
      else
        PaintVectorLayer(TP4G_VectorLayer(Layers[LayerIdx]));
      end;
    end;
end;


procedure Register;
begin
  RegisterComponents('Pascal for GIS', [TP4G_Map]);
end;


{$IFDEF LCL}
initialization
{$I p4g_ctrls.lrs}  //Glyphs courtesy of DotSpatial project (http://dotspatial.codeplex.com)
{$ENDIF}

end.
